package io.tswf.minecraft.blockventure.core.transmutation.transmute

import groovy.transform.CompileStatic
import groovy.util.logging.Slf4j
import io.tswf.minecraft.blockventure.core.api.world.chunk.PersistedChunkCoord
import io.tswf.minecraft.blockventure.core.world.chunk.ChunkPersistService
import io.tswf.transmutations.common.annotation.ClassReference
import io.tswf.transmutations.common.annotation.Constant
import io.tswf.transmutations.common.annotation.FieldDescription
import io.tswf.transmutations.common.annotation.MethodDescription
import io.tswf.transmutations.transmute.annotation.AtHead
import io.tswf.transmutations.transmute.annotation.FieldValue
import io.tswf.transmutations.transmute.annotation.Offset
import io.tswf.transmutations.transmute.annotation.ReturningConstantIfTrue
import io.tswf.transmutations.transmute.annotation.ReturningVoidIfTrue
import io.tswf.transmutations.transmute.annotation.Transmute
import io.tswf.transmutations.transmute.injection.contexts.InjectionOffset
import io.tswf.transmutations.transmute.meta.MissingTargetStrategy
import it.unimi.dsi.fastutil.longs.Long2ObjectLinkedOpenHashMap
import net.minecraft.client.world.ClientChunkManager
import net.minecraft.client.world.ClientWorld
import net.minecraft.server.world.ChunkHolder
import net.minecraft.server.world.ServerWorld
import net.minecraft.server.world.ThreadedAnvilChunkStorage
import net.minecraft.util.math.ChunkPos

import java.util.function.BooleanSupplier

@Slf4j
@CompileStatic
class TransmutationThreadedAnvilChunkStorage {

    @Transmute(
            targetClass = @ClassReference(ThreadedAnvilChunkStorage),
            targetMethod = @MethodDescription(
                    name = "unloadChunks",
                    parameterTypeClasses = [ BooleanSupplier ],
                    returnTypeClass = ClassReference.VOID
            ),
            onMissing = MissingTargetStrategy.THROW
    )
    @Offset(InjectionOffset.BEFORE)
    @AtHead
    static void removePersistedChunksFromUnloadQueue(

            BooleanSupplier shouldKeepTicking,

            @FieldValue(@FieldDescription(name = "chunksToUnload", type = @ClassReference(Long2ObjectLinkedOpenHashMap)))
            Long2ObjectLinkedOpenHashMap<ChunkHolder> chunksToUnload,

            @FieldValue(@FieldDescription(name = "world", type = @ClassReference(ServerWorld)))
            ServerWorld world
    ) {
        def chunkCoord = new PersistedChunkCoord()
        chunksToUnload.removeAll {
            chunkCoord.x = it.value.pos.x
            chunkCoord.z = it.value.pos.z

            def result = ChunkPersistService.isChunkPersisted(world.dimensionKey.value, chunkCoord)

            if (result) {
                log.info("persisted chunk ${chunkCoord}")
            }

            return result
        }
    }
    @Transmute(
            targetClass = @ClassReference(ThreadedAnvilChunkStorage),
            targetMethod = @MethodDescription(
                    name = "shouldTick",
                    parameterTypeClasses = [ ChunkPos ],
                    returnTypeClass = ClassReference.BOOLEAN
            ),
            onMissing = MissingTargetStrategy.THROW
    )
    @AtHead
    @ReturningConstantIfTrue(@Constant(booleanValue = true))
    static boolean shouldTick(

            ChunkPos pos,

            @FieldValue(@FieldDescription(name = "world", type = @ClassReference(ServerWorld)))
            ServerWorld world
    ) {
        ChunkPersistService.isChunkPersisted(world.dimensionKey.value, new PersistedChunkCoord(x: pos.x, z: pos.z))
    }
}

@CompileStatic
class TransmutationClientChunkManager {

    @Transmute(
            targetClass = @ClassReference(ClientChunkManager),
            targetMethod = @MethodDescription(
                    name = "unload",
                    parameterTypeClasses = [ ClassReference.INT, ClassReference.INT ],
                    returnTypeClass = ClassReference.VOID
            ),
            onMissing = MissingTargetStrategy.THROW
    )
    @AtHead
    @Offset(InjectionOffset.BEFORE)
    @ReturningVoidIfTrue
    static boolean cancelUnloadClientChunk(

            int chunkX,
            int chunkY,

            @FieldValue(@FieldDescription(name = "world", type = @ClassReference(ClientWorld)))
            ClientWorld clientWorld
    ) {
        return ChunkPersistService.isChunkPersisted(clientWorld.dimensionKey.value, new PersistedChunkCoord(x: chunkX, z: chunkY))
    }
}
